//
// C Implementation: calib
//
// Description: Calibration for data streams in owwlog
//
//
// Author: Simon Melhuish <simon@melhuish.info>, (C) 2004
//
// Copyright: See COPYING file that comes with this distribution
//
//

// Structure for stream calibration

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

//#ifdef _WIN32
//#  include "win32config.h"
//#endif

#ifdef HAVE_LIBGSL

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "calib.h"
#include "owwl.h"
#include "logger.h"

#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>

#ifdef _WIN32
#define DIRSEP "\\"
#else
#define DIRSEP "/"
#endif

typedef struct streamcal_ {
  //int id ;
  //enum calib_state_enum cal_state ;
  gsl_interp_accel *acc ;
  gsl_spline *spline ;
  //struct streamcal_ *next ;
} streamcal ;

typedef struct cal_pair_ {
  double x ;
  double y ;
} cal_pair ;

/* Messages, warnings and errors */
void owwlog_message(char* format, ...) ;
void owwlog_warning(char* format, ...) ;
void owwlog_error(char* format, ...) ;

void reduce_name(char *name) ;

//streamcal *cal_list = NULL ;

// Read 0 to size doubles from a line, returning number of doubles read
static int
read_numbers(char *line, double vals[], int size)
{
  int i ;
  char *p ;
  
  p = &line[0] ;
  
  if (line[0] == '#') return 0 ;
  
  for (i=0; i<size; ++i)
  {
    char *endptr ;
    double val ;
    
    val = strtod(p, &endptr) ;
    if (p == endptr) // No conversion?
      return i ;
    
    if (vals != NULL)
      vals[i] = val ;
    
    p = endptr ;
  }
  
  return i ;
}

static int
compare_cals(const void *c1, const void *c2)
{
  if (((cal_pair *)c1)->x < ((cal_pair *)c2)->x) return -1 ;
  if (((cal_pair *)c1)->x > ((cal_pair *)c2)->x) return  1 ;
  return 0 ;
}

// Load a csv file in two columns into x and y arrays
static int
calib_load(char *file_name, double **xp, double **yp, int *np)
{
  FILE *fp ;
  char line[256] ;
  int count = 0 ;
  double vals[2] ;
  cal_pair *cals ;
  int i ;
  
  //owwlog_message("Opening calib file \"%s\"\n", file_name) ;
  
  fp = fopen(file_name, "r") ;
  
  if (fp == NULL)
    owwlog_error("Unable to open calib file \"%s\"\n", file_name) ;
  
  // Load each line to count values
  while (!feof(fp))
  {
    fgets(line, sizeof(line), fp) ;
    if (read_numbers(line, vals, 2) == 2)
      ++count ;
  }
  
  //owwlog_message("Allocating calib data for %d lines\n", count) ;
  
  // Allocate arrays
  
  *xp = (double *) calloc(count, sizeof(double)) ;
  *yp = (double *) calloc(count, sizeof(double)) ;
  *np = count ;
  cals = (cal_pair *) calloc(count, sizeof(cal_pair)) ;
  
  // Rewind to start of file
  
  rewind(fp) ;
  count = 0 ;
  
  // Load values into arrays
  while (!feof(fp) && (count < *np))
  {
    fgets(line, sizeof(line), fp) ;
    if (read_numbers(line, vals, 2) == 2)
    {
      //owwlog_message("%g, %g\n", vals[0], vals[1]);
      //(*xp)[*np-count-1] = vals[1] ;
      //(*yp)[*np-count-1] = vals[0] ;
      cals[count].x = vals[1] ;
      cals[count].y = vals[0] ;
      ++count ;
    }
  }
  
  // Sort the cal values by x
  
  qsort(cals, count, sizeof(cal_pair), compare_cals) ;
  
  for (i=0; i<count; ++i)
  {
    //owwlog_message("%g, %g\n", cals[i].x, cals[i].y) ;
    (*xp)[i] = cals[i].x ;
    (*yp)[i] = cals[i].y ;
  }
  
  free(cals) ;
  
  owwlog_message("%s: Read %d lines of calibration data\n", file_name, count) ;
  //owwlog_message("T range %g to %g\n", (*yp)[0], (*yp)[count-1]) ;
  //owwlog_message("Value range %g to %g\n", (*xp)[0], (*xp)[count-1]) ;
  
  fclose(fp) ;
  
  return 0 ;
}

/* Check a stream for possible calibration file, and set up if found */
void *
calib_new(owwl_conn *conn, owwl_data *data, char *caldir)
{
  char *full_file_name, stream_name[128] ;
  int name_length ;
  streamcal *cal ;
  double *x ;
  double *y ;
  int n ;
  
  // Get the reduced form of the stream name
  owwl_name(data, stream_name, 128, &name_length, 0) ;
  reduce_name(stream_name) ;
  
  full_file_name = (char *) calloc(1, (name_length + strlen(caldir) + strlen("DIRSEP") + 1) * sizeof(char)) ;
  strcpy(full_file_name, caldir) ;
  strcat(full_file_name, DIRSEP) ;
  strcat(full_file_name, stream_name) ;  
  
  //free(stream_name) ;
  
  if (0 != access(full_file_name, F_OK))
  {
    owwlog_message("No calibration file \"%s\"\n", full_file_name) ;
    free(full_file_name) ;
    return NULL ; // No calib file
  }
  
  // Found a calib file
  
  // Create cal object
  cal = (streamcal *) calloc(1, sizeof(streamcal)) ;
  
  // Load calib data
  calib_load(full_file_name, &x, &y, &n) ;
  
  // Free filename string
  free(full_file_name) ;
  
  if (n <= 0)
  {
    free((void *) cal) ;
    return NULL ;
  }
  
  
  // Set up cal object
  cal->acc = gsl_interp_accel_alloc() ;
  cal->spline = gsl_spline_alloc(gsl_interp_akima, n) ;
  gsl_spline_init(cal->spline, x, y, n) ;
  
  // Return cal object
  return (void *) cal ;
}

static double grt_nearest_I(double I)
{
  int i ;
  double clos[] = {
  0.1, 0.25, 0.5, 
  1.0, 2.5, 5.0, 
  10.0, 25.0, 50.0, 
  100.0, 250.0, 500.0, 
  1000.0, 2500.0, 5000.0, 
  10000.0, 25000.0, 50000.0, 
  100000.0, 250000.0, 500000.0, 
  1000000.0, 2500000.0, 5000000.0, 
  } ;
  double mult[] = {
  10.0, 4.0, 2.0, 
  1.0, 0.4, 0.2,  
  0.1, 0.04, 0.02,  
  0.01, 0.004, 0.002, 
  0.001, 0.0004, 0.0002, 
  0.0001, 0.00004, 0.00002, 
  0.00001, 0.000004, 0.000002, 
  0.000001, 0.0000004, 0.0000002, 
  } ;
  
  for (i=0; i<5; ++i)
  {
    double Is ;
    
    Is = I * mult[i] ;
    
    if ((Is > 0.9) && (Is < 1.1)) return clos[i] ;
  }
  
  return I ;
}

int
calib_eval(owwl_conn *conn, owwl_data *data)
{
  // Perform interpolated calibration on a data stream
  
  streamcal *cal ;
  
  if (!data->user_data) return 0 ; // Nothing to do
  if (!((Logger *) data->user_data)->calib) return 0 ; // Nothing to do
  
  cal = (streamcal *) ((Logger *) data->user_data)->calib ;
  
  // Action depends on device type
  switch (data->device_type)
  {
    case OwwlDev_Diode:
    {
      double T ;
      
      //owwlog_message("Diode %d, V = %f\n", data->index, data->device_data.diode.V) ;
      
      T = gsl_spline_eval(cal->spline, data->device_data.diode.V, cal->acc) ;
      
      //owwlog_message("Diode %d, T = %f\n", data->index, T) ;
      
      owwl_data_in(conn, OwwlDev_Cryo, OwwlCryo_Diode, data->index, T) ;
      break ;
    }
      
    case OwwlDev_Grt:
    {
      double T, R ;
      
      R = 1000.0 * sqrt(data->device_data.grt.V * data->device_data.grt.V 
                      + data->device_data.grt.Vq * data->device_data.grt.Vq)
                      / grt_nearest_I(data->device_data.grt.I) ;
      
      //printf("%g > %g\n", data->device_data.grt.I, grt_nearest_I(data->device_data.grt.I)) ;
      
      T = gsl_spline_eval(cal->spline, R, cal->acc) ;
      
      owwl_data_in(conn, OwwlDev_Cryo, OwwlCryo_Grt, data->index, T) ;
      break ;
    }
      
    case OwwlDev_Resistance:
    {
      double T ;
      
      T = gsl_spline_eval(cal->spline, data->device_data.resistance.R, cal->acc) ;
      
      //printf("%g -> %g\n", data->device_data.resistance.R, T);
      
      owwl_data_in(conn, OwwlDev_Cryo, OwwlCryo_Grt, data->index, T) ;
      break ;
    }
  }
  
  return 0 ;
}
#endif
