/* owwl.h */


/*! \file owwl.h
    \brief Headers for oww_trx client library 
    
*/

/* Dr. Simon J. Melhuish */

/* Types for sensor data */

#ifndef OWWL_H
#define OWWL_H 1

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* ifdef HAVE_CONFIG_H */

#include <time.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif /* ifdef HAVE_SYS_TYPES_H */

#ifdef HAVE_INTTYPES_H
# include <inttypes.h>
#endif

#ifndef SOCKET
#define SOCKET int
#endif

/* #define OWWL_DEBUG 1 */
/** \addtogroup MiscDef_ Miscellaneous Definitions 
\{ */
#define OWW_PROTO_VERSION 101L /**< Owwl Protocol Version Number */
/** \} */

/**
 * \addtogroup Units_ Data Units
 * \{
 */

/**
 * \addtogroup UnitEnum_ Unit Class Enumeration
 * \{
 */

/** \brief An enumeration used to specify the class of unit associated with a data stream */
enum owwl_unit_class_enum {
  owwl_unit_class_unknown = -1,      /**< Unknown unit class */
  owwl_unit_class_dimensionless = 0, /**< For dimensionless data streams - no unit */
  owwl_unit_class_time,              /**< A time value */
  owwl_unit_class_count,             /**< A count */
  owwl_unit_class_count_rate,        /**< A count rate */
  owwl_unit_class_length,            /**< A length, e.g. inches */
  owwl_unit_class_length_rate,       /**< A length rate, e.g. inches per hour */
  owwl_unit_class_speed,             /**< A speed, e.g. mph */
  owwl_unit_class_temperature,       /**< A temperature, e.g. degrees C */
  owwl_unit_class_pressure,          /**< A pressure, e.g. mBar */
  owwl_unit_class_direction,         /**< A direction, e.g. as a bearing from north, in degrees */
  owwl_unit_class_humidity,          /**< A humidty */
  owwl_unit_class_uvi,               /**< A UV index */
  owwl_unit_class_voltage,           /**< A Voltage */
  owwl_unit_class_mvoltage,          /**< A Voltage in milli Volts */
  owwl_unit_class_muvoltage,         /**< A Voltage in micro volts */
  owwl_unit_class_nanocurrent,       /**< A current in nano amps */
  owwl_unit_class_cryo,              /**< A cryogenic temperature */
  owwl_unit_class_resistance,        /**< A resistance (from an RTD) */
  owwl_unit_class_mamps,             /**< A current, in mA */
  owwl_unit_class_coulombs,          /**< A charge, in C */
  owwl_unit_class_mvs,               /**< ADC accumulator value - mV.s */
  OWWL_UNIT_CLASS_LIMIT
} ;
/**
 * \}
 */


/* #defines for units */

/** \brief Codes for unit choice
 *
 * These defines give the codes to specify a unit to use for a conversion.
 * \addtogroup UnitDefines_ Unit Numbers
 * \{
 */

/**
 * \addtogroup MainUnits_ The main unit groups
 * \{
 */

#define OwwlUnit_Metric   0 /**< "Metric" units */
#define OwwlUnit_Imperial 1 /**< "Imperial" units */

#define OwwlUnit_Alt1     2 /**< Usually alternative metric */
#define OwwlUnit_Alt2     3 /**< Usually alternative imperial */
/**
 * \}
 */

/**
  * \addtogroup TempUnits_ Temperatures
  * \{
  */
#define OwwlUnit_Celcius    OwwlUnit_Metric
#define OwwlUnit_Fahrenheit OwwlUnit_Imperial
#define OwwlUnit_Kelvin     OwwlUnit_Alt1
#define OwwlUnit_Rankine    OwwlUnit_Alt2
/**
 * \}
 */

/**
  * \addtogroup CryoUnits_ Cryogenic Temperatures
  * \{
  */
#define OwwlUnit_CryoKelvin     OwwlUnit_Metric
#define OwwlUnit_CryoRankine    OwwlUnit_Imperial
#define OwwlUnit_CryoCelcius    OwwlUnit_Alt1
#define OwwlUnit_CryoFahrenheit OwwlUnit_Alt2
/**
 * \}
 */

/**
  * \addtogroup LengthUnits_ Lengths
  * \{
  */
#define OwwlUnit_Mm         OwwlUnit_Metric   /**< mm */
#define OwwlUnit_Inch       OwwlUnit_Imperial /**< inches */
/**
 * \}
 */

/**
  * \addtogroup SpeedUnits_ Speed
  * \{
  */
#define OwwlUnit_Kph        OwwlUnit_Metric   /**< km / hr */
#define OwwlUnit_Mph        OwwlUnit_Imperial /**< miles / hr */
#define OwwlUnit_Mps        OwwlUnit_Alt1     /**< m / s */
#define OwwlUnit_Knot       OwwlUnit_Alt2     /**< Knots */
/**
 * \}
 */

/**
  * \addtogroup UVUnits_ UV
  * \{
  */
#define OwwlUnit_UVI        OwwlUnit_Metric   /**< UV index */
#define OwwlUnit_mWpersqm   OwwlUnit_Alt1     /**< mW / m^2 */
/**
 * \}
 */

/**
  * \addtogroup VUnits_ V
  * \{
  */
#define OwwlUnit_V    OwwlUnit_Metric   /**< Volts */
#define OwwlUnit_mV   OwwlUnit_Imperial /**< milli Volts */
/**
 * \}
 */

/**
  * \addtogroup VUnits_ mV
  * \{
  */
#define OwwlUnit_mV_mV    OwwlUnit_Metric   /**< milli Volts */
#define OwwlUnit_mV_uV    OwwlUnit_Imperial /**< micro Volts */
/**
 * \}
 */

/**
  * \addtogroup VsUnits_ mVs
  * \{
  */
#define OwwlUnit_mVs_mVs    OwwlUnit_Metric   /**< milli Volt seconds */
#define OwwlUnit_mVs_mVhr   OwwlUnit_Imperial /**< micro Volt hours */
#define OwwlUnit_mVs_Vs     OwwlUnit_Alt1     /**< Volt seconds */
#define OwwlUnit_mVs_Vhr    OwwlUnit_Alt2     /**< Volt hours */
/**
 * \}
 */

/**
  * \addtogroup PressureUnits_ Pressure
  * \{
  */
#define OwwlUnit_Mbar       OwwlUnit_Metric   /**< mBar */
#define OwwlUnit_InchHg     OwwlUnit_Imperial /**< Inches of mercury */
#define OwwlUnit_Kpa        OwwlUnit_Alt1     /**< kPascal */
#define OwwlUnit_Atm        OwwlUnit_Alt2     /**< Standard atmospheres */
/**
 * \}
 */

/**
  * \addtogroup WinUnits_ Wind direction
  * \{
  */
#define OwwlUnit_Bearing OwwlUnit_Metric     /**< Bearing in degrees clockwise from north */
#define OwwlUnit_Point16 OwwlUnit_Imperial   /**< Bearing as a point number */
#define OwwlUnit_Name    OwwlUnit_Alt1       /**< 0 - North, clockwise to 15 - NNW */
#define OwwlUnit_Radians OwwlUnit_Alt2       /**< Bearing in radians clockwise from north */
/**
 * \}
 */

/**
  * \addtogroup RateUnits_ Rates
  * \{
  */
#define OwwlUnit_CountsPerHour       OwwlUnit_Imperial   /**< counts per hour */
#define OwwlUnit_CountsPerSec        OwwlUnit_Metric /**< counts per second */
/**
 * \}
 */

/**
 * \}
 */

/**
 * \}
 */

/** \addtogroup Devices_ Devices
\{ */

/** \addtogroup DevTypes_ Device Types
\{ */

/**
  * \brief Enumeration for different types of device
  *
  * Each different kind of device has a primary device type.
  * This is given according to the owwl_device_type_enum enumeration, in
  * the device_type member of an owwl_data
  *
  * This is not quite the same as owwl_unit_class_enum because different
  * types of device may share a common unit class.
  */

enum owwl_device_type_enum {
  OwwlDev_Any = -1,        /**< For use in a search - any device type will match */
  OwwlDev_Temperature = 0, /**< A thermometer */
  OwwlDev_Barometer,       /**< A Barometer */
  OwwlDev_Humidity,        /**< A hygrometer (humidity device) */
  OwwlDev_Wind,            /**< A wind sensor (direction and speed) */
  OwwlDev_Rain,            /**< A rain gauge */
  OwwlDev_Gpc,             /**< A general-purpose counter */
  OwwlDev_Solar,           /**< A solar sensor */
  OwwlDev_Uvi,             /**< A UV sensor */
  OwwlDev_Adc,             /**< An ADC channel */
  OwwlDev_Psb = 64,        /**< A polarization-sensitive bolometer */
  OwwlDev_Diode,           /**< A diode thermometer */
  OwwlDev_Grt,             /**< A GRT */
  OwwlDev_Cryo,            /**< A calibrated cryogenic thermometer */
  OwwlDev_Heater,          /**< A heater voltage */
  OwwlDev_Resistance,      /**< A resistance device */
  OwwlDev_Invalid
} ;

/** \brief Some device types can be split by different sub-types
  * \addtogroup DevSubTypes_ Device Sub-Types
\{ */

#define OwwlSubtype_Any -1    /**< We want to select ignoring sub-type */
#define OwwlSubtype_Default 0 /**< The default sub-type */

/**
 * \brief Enum for thermometer sub-types
 *
 * Temperatures may come from several sources
 */

enum owwl_temperature_subtype_enum {
  OwwlTemp_Any = OwwlSubtype_Any, /**< Used when we want to select any sub-type */
  OwwlTemp_Thermometer = OwwlSubtype_Default, /**< The default sub-type - normal thermometer */
  OwwlTemp_Humidity, /**< Temperature from hygrometers (humidity sensors) */
  OwwlTemp_Barometer,/**< Temperature from barometers */
  OwwlTemp_UVI,      /**< Temperature from UVI sensors */
  OwwlTemp_Solar,    /**< Temperature from solar sensors */
  OwwlTemp_Adc,      /**< Temperature from ADC channels */
  OwwlTemp_Thrmcpl,  /**< Temperature from Thermocouple (compensated) */
  OwwlTemp_ThrmcplCJ /**< Temperature from Thermocouple cold junction */
} ;
/** \} */

/**
 * \brief Structure to hold thermometer data 
 *
 */
typedef struct {
  float T ;          /**< Temperature in C */
  float padding[9] ;
} owwl_data_temperature ;

/**
 * \brief Structure to hold barometer data 
 *
 */
typedef struct {
  float bp ;          /**< Barometric pressure in mBar */
  float padding[9] ;
} owwl_data_barometer ;

/**
 * \brief Structure to hold hygrometer data
 *
 */
typedef struct {
  float RH ;          /**< Realtive humidity */
  float padding[9];
} owwl_data_humidity ;

/**
 * \brief Structure to hold solar sensor data
 *
 */
typedef struct {
  float solar ;       /**< Solar radiation */
  float padding[9];
} owwl_data_solar ;

/**
 * \brief Structure to hold UVI sensor data
 *
 */
typedef struct {
  float uvi ;       /**< UV index */
  float padding[9];
} owwl_data_uvi ;

/**
 * \brief Structure to hold ADC channel data
 *
 */
typedef struct {
  float V ; /**< Voltage (V) */
  float I ; /**< Current (mA) */
  float Q ; /**< Accumulated charge (Coulomb) */
  float padding[7];
} owwl_data_adc ;

/**
 * \brief Structure to hold wind data 
 *
 */
typedef struct {
  float speed ;      /**< Wind speed in m/s */
  float gust ;       /**< Gust (max) speed in m/s */
  float bearing ;    /**< Bearing clockwise from North in degrees */
  float padding[7] ;
} owwl_data_wind ;

/**
 * \brief Structure to hold rain data 
 *
 */
typedef struct {
  unsigned int rain_count ; /**< Monotonic rain count - counts */
  time_t rain_reset_time ;  /**< Time of last reset */
  float  rain_since_reset ; /**< Rain in mm since last rest */
  float  rain_rate ;        /**< Rain rate in mm/hr */
  float  padding[6] ;
} owwl_data_rain ;

/**
 * \brief Structure to hold general-purpose counter data 
 *
 */
typedef struct {
  unsigned int gpc_count ;/**< GPC monotonic counter value - counts */
  time_t gpc_reset_time ; /**< Time of last GPC reset */
  float  gpc_calibrated ; /**< GPC count * calibration factor since reset */
  float  gpc_rate ;       /**< GPC count rate - counts / second */
  float  gpc_delta ;      /**< Change in gpc_calibrated since last reading */
  float  padding[5] ;
} owwl_data_gpc ;

/**
 * \brief Structure to hold PSB device data 
 *
 */
typedef struct {
  int  psb_a ;            /**< PSB channel A raw adc value */
  int  psb_b ;            /**< PSB channel B raw adc value */
  float  padding[8] ;
} owwl_data_psb ;

/**
 * \brief Structure to hold diode thermometer data 
 *
 */
typedef struct {
  float  V ;            /**< Diode voltage value */
  float  padding[9] ;
} owwl_data_diode ;

/**
 * \brief Structure to hold GRT device data 
 *
 */
typedef struct {
  float  V ;             /**< GRT V (in-phase) voltage value */
  float  Vq ;            /**< GRT Vq (quadrature) voltage value */
  float  I ;             /**< GRT current value */
  float  padding[7] ;
} owwl_data_grt ;

/**
 * \brief Structure to hold calibrated cryogenic thermometer data 
 *
 */
typedef struct {
  float T ;          /**< Temperature in K */
  float padding[9] ;
} owwl_data_cryo ;

/**
 * \brief Structure to hold heater voltage data 
 *
 */
typedef struct {
  float  V ;            /**< Heater voltage value */
  float mA ;            /**< Heater Isense (mA) */
  int   haveI ;         /**< 1 if we have Isense data, otherwise 0 */
  float  padding[7] ;
} owwl_data_heater ;

/**
 * \brief Structure to hold resistance data 
 *
 */
typedef struct {
	float  R ;            /**< Resistance value in Ohms */
	float  padding[9] ;
} owwl_data_resistance ;

/**
 * \brief Enum for calibrated cryogenic thermometer sub-types
 *
 * Temperatures may come from several sources
 */

enum owwl_cryo_subtype_enum {
  OwwlCryo_Any = OwwlSubtype_Any, /**< Used when we want to select any sub-type */
  OwwlCryo_Thermometer = OwwlSubtype_Default, /**< The default sub-type - normal thermometer */
  OwwlCryo_Diode, /**< Temperature from diodes */
  OwwlCryo_Grt /**< Temperature from GRTs */
} ;
/** \} */

/**
 * \brief Structure to hold device data - general floats
 *
 */
typedef struct {
  float floats[10] ;
} owwl_data_floats ;
/** \} */

/**
 * \brief A union of all possible device data structures
 *
 */
typedef union {
    owwl_data_temperature temperature ; /**< Temperature data */
    owwl_data_humidity    humidity ;    /**< Humidity data */
    owwl_data_wind        wind ;        /**< Wind data */
    owwl_data_rain        rain ;        /**< Rain data */
    owwl_data_gpc         gpc ;         /**< General-purpose counter data */
    owwl_data_barometer   barom ;       /**< Barometer data */
    owwl_data_solar       solar ;       /**< Solar data */
    owwl_data_uvi         uvi ;         /**< UVI data */
    owwl_data_adc         adc ;         /**< ADC data */
    owwl_data_psb         psb ;         /**< Polarization-sensitive bolometer data */
    owwl_data_diode       diode ;       /**< Diode thermometer data */
    owwl_data_grt         grt ;         /**< GRT data */
    owwl_data_cryo        cryo ;        /**< Cryo data */
    owwl_data_heater      heater ;      /**< Heater voltage data */
    owwl_data_resistance  resistance ;  /**< Resistance (of an RTD) */
    owwl_data_floats      floats ;      /**< Array of floats */
} device_data_union ; 

/**
 * \brief Structure used to hold a data item.
 *
 * Data are stored in an array of "owwl_data". 
 * Each item contains a union of 
 * data structures for the various data types.
 * Functions to convert data to strings or doubles are provided.
 */
typedef struct owwl_data_ {
  int device_type ; /**< The type of device held in the struct */
  device_data_union device_data ;   /**< A union of all possible device data structures */
  int device_subtype     ; /**< Some devices have sub-types. For example, a
                                temperature might come from a hygrometer */
  int index ; /**< The index value, running from 0, e.g. thermometer 0, &c. */
  int in_use ; /**< Is this an up-to-date in-use data item? */
  void *user_data ; /**< For unspecified use by user program */
  /** Function to convert one arg of the data item to a string, with given unit type and precision */
  char * (*str)(struct owwl_data_ *self, char *buffer, int bufflen, int unit_type, int precision, int arg) ;
  /** Function to convert one arg of the data item to a double, with given unit type */
  double (*val)(struct owwl_data_ *self, int unit_type, int arg) ;
  /* char * (*unit)(struct owwl_data_ *self, int unit_type, int arg) ; */
  /* int (*next_arg)(struct owwl_data_ *self, int arg) ; */
} owwl_data ;
/** \} */

/**
  * \addtogroup OwwlConn_ Owwl Connexion
\{ */

/** \brief Structures pertaining to server operation
  * \addtogroup Server_ Server Structures
\{ */

/**
 * \brief Structure used to hold client connexion data
 *
 * An owwl_server may point to an array of owwl_client, with "first_client".
 * These form a doubly-linked list.
 * Each owwl_client gives a socket fd for one connected client, to which 
 * data should be sent.
 */
typedef struct owwl_client_ {
  struct owwl_client_ *prev ; /**< Link back */
  struct owwl_client_ *next ; /**< Link forward */
  SOCKET socket ;                     /**< A client socket */
} owwl_client ;

/**
 * \brief Structure used to hold server data
 *
 * An owwl_conn may point to an array of owwl_server, with "server".
 * This is a singly-linked list of sockets and clients.
 * The sockets listen for incoming client connexions.
 * first_client points to the start of a doubly-linked list of clients.
 */
typedef struct owwl_server_ {
  SOCKET socket_listen ;              /**< The listening socket */
  owwl_client *first_client ;      /**< Start of doubly-linked list of client sockets */
} owwl_server ;
/** \} */

struct readstr {
  int data_size ;
  int data_to_get ;
  uint32_t data_word_n[16] ;
  int state ;
  uint32_t head_word;
  int head_bytes_read;
} ;

/**
 * \brief Structure used to hold connexion data
 *
 * Data source may be local (server operation) or remote (client operation)
 */
typedef struct {
  SOCKET socket ;       /**< fd of socket used to receive data from server */
  int read_result ;     /**< Not used */
  time_t data_time ;    /**< Time-stamp on latest data - unix time */
  long data_time_usec;  /**< Time-stamp on latest data - micro seconds to add to data_time */
  int interval ;        /**< Time interval between samples - seconds */
  float longitude ;     /**< The longitude of the server */
  float latitude ;      /**< The latitude of the server */
  owwl_data *data ;     /**< Pointer to first data item */
  int data_count ;      /**< The number of items in data */
  void *user_data ;     /**< Arbitrary pointer for user */
  owwl_server *server ; /**< Pointer to first server */
  int nserver ;         /**< Number of items in server */
  struct readstr rd ;   /**< Structure for owwl_read */
} owwl_conn ;

/**
 * \brief Structure used to build transmit (server) data stream
 */
typedef struct {
  uint32_t *buffer ; /**< Buffer of words */
  int next ;       /**< Next word in buffer to write to */
  int size ;       /**< Total number of words malloced */
} owwl_buffer ;

#define OwwlBuffInc 16 /**< Step size for allocating buffer memory */
#define Owwl_Buffer_Init {(uint32_t *) NULL, 0, 0}

/** Function type for owwl_foreach */

typedef int (*owwl_func) (owwl_conn *conn, owwl_data *data, void *user_data) ;

/** 
  * \brief The values returned by owwl_read()
  */

enum owwl_read_result {
  Owwl_Read_Disconnect = -2,      /**< The remote server has disconnected */
  Owwl_Read_Error = -1,           /**< An error occurred */
  Owwl_Read_Read_And_Decoded = 0, /**< The whole message has been read and decoded Ok */
  Owwl_Read_Again                 /**< More data need to be read */
} ;
/** \} */

/** 
  * \brief The values returned by owwl_decode()
  */

enum owwl_decode_result {
  Owwl_Decode_Error = -1,   /**< An error occurred */
  Owwl_Decode_Not_Done = 0, /**< More data need to be decoded */
  Owwl_Decode_Done          /**< The message decoded Ok */
} ;
/** \} */

/* Header Word Codes */
/** \addtogroup MiscDef_ Miscellaneous Definitions 
\{ */
/**
 * \addtogroup HeaderWords_ Header Word Codes
 * \{
 */
#define OWW_TRX_END              0L /**< To mark end of data */
#define OWW_TRX_TIME     0x1000000L /**< Times */
#define OWW_TRX_WIND     0x2000000L /**< Wind direction, speed, gust and max */
#define OWW_TRX_RAIN     0x3000000L /**< Rain mono, rain, since, rate, in hour */
#define OWW_TRX_T        0x4000000L /**< T (C) * 100 or Trh (C) * 100 */
#define OWW_TRX_RH       0x6000000L /**< RH (%) * 100 */
#define OWW_TRX_BP       0x7000000L /**< BP (mB) * 10 */
#define OWW_TRX_SOLAR    0xb000000L /**< Solar (arbitrary units) */
#define OWW_TRX_UVI      0xc000000L /**< UVI (index value = weighted flux density integral / (10 mW/m^2) */
#define OWW_TRX_ADC      0xd000000L /**< ADC chanel - V (V), I (mA) and Q (C) */
#define OWW_TRX_PSB     0x10000000L /**< Polarization-sensitive bolometer - A and B */
#define OWW_TRX_DIODE   0x11000000L /**< Diode thermometer voltage */
#define OWW_TRX_GRT     0x12000000L /**< GRT V, Vq, and I */
#define OWW_TRX_CRYO    0x13000000L /**< Calibrated cryogenic T (from Diode or GRT) */
#define OWW_TRX_HEATER  0x14000000L /**< Heater voltage */
#define OWW_TRX_RESISTANCE  0x15000000L /**< Resistance (of an RTD) */
#define OWW_TRX_UPDT    0x80000000L /**< Update interval only */
#define OWW_TRX_GPC     0x90000000L /**< gpc, gpcdelta, gpcmono, gpcrate */
#define OWW_TRX_LOC     0xa0000000L /**< location */
#define OWW_TRX_HEAD    0xff000000L /**< To mark header of normal data */

#define OWW_TRX_SUB_TRH    0x10000L /**< Trh subtype of T */
#define OWW_TRX_SUB_TB     0x20000L /**< Tb subtype of T */
#define OWW_TRX_SUB_TUV    0x30000L /**< Tuv subtype of T */
#define OWW_TRX_SUB_TSOL   0x40000L /**< Tsol subtype of T */
#define OWW_TRX_SUB_TADC   0x50000L /**< Tadc subtype of T */
#define OWW_TRX_SUB_TTC    0x60000L /**< Ttc subtype of T */
#define OWW_TRX_SUB_TCJ    0x70000L /**< Tcj subtype of T */

#define OWW_TRX_SUB_TCD    0x10000L /**< Tdiode subtype of Tcryo */
#define OWW_TRX_SUB_TCG    0x20000L /**< Tgrt subtype of Tcryo */

#define OWW_TRX_SERIAL       0x100L /**< For serial number byte */

#define OWW_TRX_SIZE             1L /**< Size byte */
/** \} */

/**
 * \addtogroup MessageTypes_ Message types
 * \{
 */
#define OWW_TRX_MSG_WSDATA 1 /**< Normal data stream update */
#define OWW_TRX_MSG_UPDT   2 /**< Update interval has changed */
#define OWW_TRX_MSG_MORT   3 /**< Server about to die */
#define OWW_TRX_MSG_ANNOUN 4 /**< Initial announcement */
/** \} */
/** \} */

/* Type numbers for devices */
/* These are derived from the values used in oww_trx */

enum owwl_device_types {
  dummy = 0
} ;

/* ----------------------------------------------------------------------------- */

/* Function prototypes */

/** \brief Create a new owwl connexion context 
  * 
  * \param socket fd of established socket to receive data from server
  * \param user_data Pointer to user data to associate with the conn
  * \return A new owwl_conn context
*/
owwl_conn *
owwl_new_conn(int socket, void *user_data) ;

/** \brief Try to read new data from server 
  * 
  * \param conn owwl_conn context for the connexion we want to read
  * \return 
  * \li owwl_read_result::Owwl_Read_Disconnect (-2) if the server disconnects
  * \li owwl_read_result::Owwl_Read_Error (-1) on any other error
  * \li owwl_read_result::Owwl_Read_Read_And_Decoded (0) for Ok
  * \li owwl_read_result::Owwl_Read_Again (1) for not ready - recv() would block
*/
int
owwl_read(owwl_conn *conn) ;

/** \brief Free up finished-with connexion context */
int
owwl_free(owwl_conn *conn) ;

/** \brief Get an integer id uniquely identifying a data stream 
  *
  * \param data Pointer to the data stream
  * \param arg Argument number
  * \return Unique identifier
  */
unsigned int
owwl_data_arg_to_id(owwl_data *data, int arg) ;

/** \brief Get a data stream pointer from a unique id value 
  *
  * \param conn owwl_conn context for the connexion
  * \param id Unique id value
  * \param data Pointer to value to receive data stream pointer
  * \param arg Pointer to integer that will receive argument number
  */
void
owwl_id_to_data_arg(owwl_conn *conn, unsigned int id, owwl_data **data, int *arg) ;


/** \brief Call a function for each data entry
  *
  * The function should return non-zero to break out of the loop 
  *
  * \param conn owwl_conn context for the connexion
  * \param func Pointer to user function to call
  * \param user_data Pointer to arbitrary space to pass to user function
  * \return
  * \li 0 on completion of full loop
  * \li -1 if the user function broke out of the loop (by returning non-zero)
  */

int
owwl_foreach(owwl_conn *conn, owwl_func func, void *user_data) ;

/** \brief Call a function for each data entry, in a set order
  *
  * The function should return non-zero to break out of the loop 
  *
  * \param conn owwl_conn context for the connexion
  * \param func Pointer to user function to call
  * \param user_data Pointer to arbitrary space to pass to user function
  * \return
  * \li 0 on completion of full loop
  * \li -1 if the user function broke out of the loop (by returning non-zero)
  */

int
owwl_foreach_ordered(owwl_conn *conn, owwl_func func, void *user_data) ;

/** \brief Get a name for a data entry
  *
  * \param data Pointer to a data stream from an owwl_conn
  * \param buffer Buffer to hold the returned name - use NULL just to get the length required
  * \param bufflen The length of the provided buffer
  * \param length Pointer to int that will take the length written to buffer on return
  * \param long_not_short Non-zero for a long version, or 0 for short
  * \return Name for the data stream
  *
  * Length of name string written to length
  */

char *
owwl_name(owwl_data *data, char *buffer, int bufflen, int *length, int long_not_short) ;

/** \brief Name for a data argument 
  *
  * \param device_type The device type
  * \param device_subtype The device sub-type
  * \param arg The argument number
  * \return Name for this type of argument
*/

char *
owwl_arg_stem(enum owwl_device_type_enum device_type, int device_subtype, int arg) ;

/** \brief Get a name for a data entry argument 
  * 
  * \param data Pointer to a data stream from an owwl_conn
  * \param buffer Buffer to hold the returned name - use NULL just to get the length required
  * \param bufflen The length of the provided buffer
  * \param length Pointer to int that will take the length written to buffer on return
  * \param long_not_short Non-zero for a long version, or 0 for short
  * \param arg The argument number from the data stream
  * \return Name for the data stream
  */

char *
owwl_arg_name(owwl_data *data, char *buffer, int bufflen, int *length, int long_not_short, int arg) ;

/** \brief Get the unit class for an arg 
  *
  * \param data Pointer to a data stream from an owwl_conn
  * \param arg The argument number from the data stream
  * \return The class of unit for this data stream argument 
  */
enum owwl_unit_class_enum owwl_unit_class(owwl_data *data, int arg) ;

/** \brief Get the unit name string for a data value 
  *
  * \param data Pointer to a data stream from an owwl_conn
  * \param unit_type The type of unit:
  * \li 0 Metric
  * \li 1 Imperial
  * \li 2 Alt Metric
  * \li 3 Alt Imperial
  * \param arg The argument number from the given unit class
  * \return The unit name string
  */
char *owwl_unit_name(owwl_data *data, int unit_type, int arg) ;

/** \brief Get the unit string for a data value 
  *
  * \param data Pointer to a data stream from an owwl_conn
  * \param unit_type The type of unit:
  * \li 0 Metric
  * \li 1 Imperial
  * \li 2 Alt Metric
  * \li 3 Alt Imperial
  * \param arg The argument number from the given unit class
  * \return The unit string
  */
char *owwl_unit_desc(owwl_data *data, int unit_type, int arg) ;

/** \brief Get the next arg number
  *
  * \param data Pointer to a data stream from an owwl_conn
  * \param arg The previous argument number
  * \return The next argument number, or -1 if the last one was already reached
  */

int owwl_next_arg(owwl_data *data, int arg) ;

/** \brief Find a device by type and index
  *
  * \param conn owwl_conn context for the connexion
  * \param device_type The device type to find, or -1 (OwwlDev_Any) for any
  * \param device_subtype The device sub-type to find, or -1 (OwwlSubtype_Any) for any
  * \param index The index number to find
  * \return Pointer to data stream
  */

owwl_data *
owwl_find(owwl_conn *conn, 
          enum owwl_device_type_enum device_type, 
          int device_subtype, 
          int index) ;

/** \brief Find the first device of the given type 
  *
  * \param conn owwl_conn context for the connexion
  * \param device_type The device type to find, or -1 (OwwlDev_Any) for any
  * \param device_subtype The device sub-type to find, or -1 (OwwlSubtype_Any) for any
  * \return Pointer to data stream
  */

owwl_data *
owwl_find_first(owwl_conn *conn, 
                enum owwl_device_type_enum device_type, 
                int device_subtype) ;

/** \brief Fill a buffer with data for transmission 
  *
  * \param conn owwl_conn context for the connexion
  * \param msg_type The type of message to build
  * \param buff The dynamic buffer in which the message will be built
  * \return 0 on success
  */
int
owwl_tx_build(owwl_conn *conn, int msg_type, owwl_buffer *buff) ;
  
/** \brief Update or add data values in a data stream 
  * 
  *  This is a convenience function for updating data streams.
  *  You could just as well access the data structs directly, which
  *  would remove the overhead of device lookup, if you retain a list of pointers.
  *  
  *  Data streams listed under \p conn are searched for a matching
  *  \p device_type, \p device_subtype and \p index.
  *  If no match is found, a new data stream is added to the list.
  *  If no match is found, a new data stream is added to the list.
  *  A variable argument list follows \p index, according to the device type.
  *
  * \param conn owwl_conn context for the connexion
  * \param device_type The device type for the devcie
  * \param device_subtype The device sub-type for the device
  * \param index The index of the device
  * \param ... Variable argument list depending on device type
  * - Temperature (#OwwlDev_Temperature)
  *   - double T
  * - Barometer (#OwwlDev_Barometer)
  *   - double bp
  * - Humidity (#OwwlDev_Humidity)
  *   - double RH
  * - Wind (#OwwlDev_Wind)
  *   - double speed
  *   - double gust
  *   - double bearing
  * - Rain (#OwwlDev_Rain)
  *   - int rain_count
  *   - time_t rain_reset_time
  *   - double rain_since_reset
  *   - double rain_rate
  * - GPC (#OwwlDev_Gpc)
  *   - int gpc_count
  *   - time_t gpc_reset_time
  *   - double gpc_calibrated
  *   - double gpc_rate
  *   - double gpc_delta
  * - PSB (#OwwlDev_Psb)
  *   - int psb_a
  *   - int psb_b
  * - Diode voltage (#OwwlDev_Diode)
  *   - double V
  * - GRT (#OwwlDev_Grt)
  *   - double V
  *   - double Vq
  *   - double I
  * - Cryogenic temperature (#OwwlDev_Cryo)
  *   - double T
  * \return 0 on success, -1 on memory exhaustion, -2 on bad device type
*/
int
owwl_data_in(
  owwl_conn *conn, 
  enum owwl_device_type_enum device_type, 
  int device_subtype,
  int index,
  ...) ;
  
  
/** \brief Create a new server from a listening socket 
  *
  * \param conn owwl_conn context for the connexion
  * \param socket The socket that is listening for connexion requests
  * \return 0 on success
  */
int
owwl_tx_new_server(owwl_conn *conn, SOCKET socket) ;

/** \brief Send contents of buffer \p buff to all clients 
  *
  * \param conn owwl_conn context for the connexion
  * \param buff The dynamic buffer in which the message was built
  * \return 0 on success
  */
int
owwl_tx(owwl_conn *conn, owwl_buffer *buff) ;

/** \brief Set the timestamp on \p conn according to the current time 
  *
  * \param conn owwl_conn context for the connexion
  */
  
void
owwl_tx_add_timestamp(owwl_conn *conn) ;

/** \brief Poll for incoming data and connexion requests 
  *
  * \param conn owwl_conn context for the connexion
  * \return 0 on success, -1 on error
  */
int
owwl_tx_poll_servers(owwl_conn *conn) ;

#endif
